function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = require('react');
var React__default = _interopDefault(React);
var reactRouterDom = require('react-router-dom');

var currentIndex = 0;
var useIndex = function useIndex() {
  var _useState = React.useState(currentIndex++),
      index = _useState[0];

  React.useEffect(function () {
    return function () {
      currentIndex--;
    };
  }, []);
  return index;
};

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

var setInitialValuesWrappers = function setInitialValuesWrappers(newInitialExtendValuesWrappers, context) {
  context.initialValuesWrappers = newInitialExtendValuesWrappers;
};
var getInitialValuesWrappers = function getInitialValuesWrappers(context) {
  return context.initialValuesWrappers;
};
var getInitialValuesWrapper = function getInitialValuesWrapper(fieldName, context) {
  return getInitialValuesWrappers(context)[fieldName];
};
var getInitialValues = function getInitialValues(context) {
  var initialValues = {};
  Object.keys(context.initialValuesWrappers).forEach(function (key) {
    var _extends2;

    var initialValuesWrapper = context.initialValuesWrappers[key];
    initialValues = _extends({}, initialValues, (_extends2 = {}, _extends2[initialValuesWrapper.name] = initialValuesWrapper.initialValue, _extends2));
  });
  return initialValues;
};
var getInitialValueByFieldName = function getInitialValueByFieldName(fieldName, context) {
  return getInitialValues(context)[fieldName];
};

var validateBoolean = function validateBoolean(value, initialValue) {
  if (value === 'true') return true;
  if (value === 'false') return false;
  if (typeof value === 'boolean') return value;
  if (initialValue.onParsedError) return initialValue.onParsedError(value);
  return initialValue.initial;
};

var validateInitial = function validateInitial(value, initialValue) {
  if (typeof value === 'object' && 'type' in value) return initialValue.initial;
  return value;
};

var validate = function validate(value, initialValue) {
  if (initialValue.validate) {
    return initialValue.validate(value);
  }

  return value;
};

var normalizeBoolean = function normalizeBoolean(value, initialValue) {
  if (Array.isArray(value)) return normalizeBoolean(value[0], initialValue);
  var newValue = validateInitial(value, initialValue);
  newValue = validateBoolean(newValue, initialValue);
  newValue = validate(newValue, initialValue);
  return newValue;
};

var normalizeJson = function normalizeJson(value, initialValue) {
  if (Array.isArray(value)) return normalizeJson(value[0], initialValue);

  if (typeof value === 'object') {
    if ('type' in value) return initialValue.initial;
    return value;
  }

  try {
    if (!isNaN(+value) || value === 'true' || value === 'false') return JSON.parse("" + value);
    return JSON.parse(value);
  } catch (e) {
    if (initialValue.onParsedError) {
      var obj = initialValue.onParsedError(value);

      try {
        var newValue = obj;
        return newValue;
      } catch (e) {
        throw new Error("onParsedError for type 'json' must returns js type, but return '" + obj + "'");
      }
    } else {
      return initialValue.initial;
    }
  }
};

var normalizeNumberArray = function normalizeNumberArray(array, initialValue) {
  var flag = null;
  var newArray = array.map(function (value) {
    if (flag) return;

    if (isNaN(+value)) {
      if (initialValue.onParsedError) {
        flag = 'parseAll';
        return;
      }

      if (initialValue.onParsedItemError) {
        var newValue = initialValue.onParsedItemError(value);

        if (isNaN(newValue) || typeof newValue !== 'number') {
          throw new Error("onParsedItemError returned '" + newValue + "' instread must returns number");
        }

        return newValue;
      } else {
        flag = 'setInitial';
        return;
      }
    }

    return +value;
  });
  if (flag === 'parseAll') return initialValue.onParsedError(array);
  if (flag === 'setInitial') return initialValue.initial;
  return newArray;
};

var normalizeBooleanArray = function normalizeBooleanArray(array, initialValue) {
  var flag = null;
  var newArray = array.map(function (value) {
    if (flag) return;

    if (value !== 'true' && value !== 'false') {
      if (initialValue.onParsedError) {
        flag = 'parseAll';
        return;
      }

      if (initialValue.onParsedItemError) {
        var newValue = initialValue.onParsedItemError(value);

        if (typeof newValue !== 'boolean') {
          throw new Error("onParsedItemError returned '" + newValue + "' instread must returns boolean");
        }

        return newValue;
      } else {
        flag = 'setInitial';
        return;
      }
    }

    return value === 'true';
  });
  if (flag === 'parseAll') return initialValue.onParsedError(array);
  if (flag === 'setInitial') return initialValue.initial;
  return newArray;
};

var normalizeArray = function normalizeArray(value, initialValue) {
  if (typeof value === 'object' && 'type' in value) return value.initial;
  var arrayValue = Array.isArray(value) ? value : [value];

  switch (initialValue.arrayType) {
    case 'number':
      return normalizeNumberArray(arrayValue, initialValue);

    case 'boolean':
      return normalizeBooleanArray(arrayValue, initialValue);

    case 'string':
      return arrayValue;

    default:
      throw new Error("We cannot support " + initialValue.arrayType + " type yet");
  }
};

var validateInitial$1 = function validateInitial(value, initialValue) {
  if (typeof value === 'object' && 'type' in value) return initialValue.initial;
  return value;
};

var validateNumber = function validateNumber(value, initialValue) {
  if (isNaN(+value)) {
    if (initialValue.onParsedError) {
      var newValue = initialValue.onParsedError(value);
      if (isNaN(newValue) || typeof newValue !== 'number') throw new Error("onParsedError for type 'number' must returns number, but return '" + newValue + "'");
      return newValue;
    } else {
      return initialValue.initial;
    }
  }

  return +value;
};

var validateEnum = function validateEnum(value, initialValue) {
  if (initialValue["enum"] !== undefined) {
    if (Array.isArray(initialValue["enum"])) {
      if (!initialValue["enum"].includes(value)) {
        return initialValue.initial;
      }
    } else {
      if (initialValue["enum"].onNonEnum !== undefined) {
        var newValue = initialValue["enum"].onNonEnum(value);

        if (!initialValue["enum"].array.includes(newValue)) {
          throw new Error("'" + newValue + "' not contains in enum array " + initialValue["enum"] + ", but you passed it");
        }

        return newValue;
      }
    }
  }

  return value;
};

var validateInteger = function validateInteger(value, initialValue) {
  if (initialValue.integer) {
    if (!Number.isInteger(value)) {
      if (typeof initialValue.integer === 'boolean') {
        return Math.floor(value);
      } else {
        if (initialValue.integer.onFloat) {
          var newValue = initialValue.integer.onFloat(value);

          if (!Number.isInteger(newValue)) {
            throw new Error("onFloat returned non-integer value, received " + newValue);
          }
        }
      }
    }
  }

  return value;
};

var validateMinMax = function validateMinMax(value, initialValue) {
  var newValue = validateMin(value, initialValue);
  newValue = validateMax(newValue, initialValue);
  return newValue;
};

var validateMin = function validateMin(value, initialValue) {
  if (initialValue.min !== undefined) {
    if (typeof initialValue.min === 'object') {
      if (value < initialValue.min.value) {
        if (initialValue.min.onLess) {
          return initialValue.min.onLess(value);
        } else {
          return initialValue.min.value;
        }
      }
    } else {
      if (value < initialValue.min) {
        return initialValue.min;
      }
    }
  }

  return value;
};

var validateMax = function validateMax(value, initialValue) {
  if (initialValue.max !== undefined) {
    if (typeof initialValue.max === 'object') {
      if (value > initialValue.max.value) {
        if (initialValue.max.onOver) {
          return initialValue.max.onOver(value);
        } else {
          return initialValue.max.value;
        }
      }
    } else {
      if (value > initialValue.max) {
        return initialValue.max;
      }
    }
  }

  return value;
};

var validate$1 = function validate(value, initialValue) {
  if (initialValue.validate) {
    return initialValue.validate(value);
  }

  return value;
};

var normalizeNumber = function normalizeNumber(value, initialValue) {
  if (Array.isArray(value)) return normalizeNumber(value[0], initialValue);
  var newValue = validateInitial$1(value, initialValue);
  newValue = validateNumber(newValue, initialValue);
  newValue = validateEnum(newValue, initialValue);
  newValue = validateInteger(newValue, initialValue);
  newValue = validateMinMax(newValue, initialValue);
  newValue = validate$1(newValue, initialValue);
  return newValue;
};

var validateInitial$2 = function validateInitial(value, initialValue) {
  if (typeof value === 'object' && 'type' in value) return initialValue.initial;
  return value;
};

var validateEnum$1 = function validateEnum(value, initialValue) {
  if (initialValue["enum"]) {
    if (Array.isArray(initialValue["enum"])) {
      if (!initialValue["enum"].includes(value)) {
        return initialValue.initial;
      }
    } else {
      if (!initialValue["enum"].array.includes(value)) {
        if (initialValue["enum"].onNonEnum) {
          var newValue = initialValue["enum"].onNonEnum(value);

          if (!initialValue["enum"].array.includes(newValue)) {
            throw new Error("'" + newValue + "' not contains in enum array " + initialValue["enum"] + ", but you passed it from onNonEnum");
          }

          return newValue;
        }
      }
    }
  }

  return value;
};

var validate$2 = function validate(value, initialValue) {
  if (initialValue.validate) {
    return initialValue.validate(value);
  }

  return value;
};

var normalizeString = function normalizeString(value, initialValue) {
  if (Array.isArray(value)) return normalizeString(value[0], initialValue);
  var newValue = validateInitial$2(value, initialValue);
  newValue = validateEnum$1(newValue, initialValue);
  newValue = validate$2(newValue, initialValue);
  return newValue;
};

var normalizeValueForUser = function normalizeValueForUser(value, initialValue, name, context) {
  if (value === undefined) return initialValue.initial;
  var wrapper = getInitialValuesWrapper(name, context);
  if (!initialValue.active) return initialValue.initial;

  if (typeof initialValue.active === 'object' && !initialValue.active.isActive) {
    if (initialValue.active.storeValue) {
      if (initialValue.active.canChangeValue) ; else {
        return wrapper.storedValue || initialValue.initial;
      }
    } else {
      return initialValue.initial;
    }
  }

  switch (initialValue.type) {
    case 'boolean':
      return normalizeBoolean(value, initialValue);

    case 'number':
      return normalizeNumber(value, initialValue);

    case 'array':
      return normalizeArray(value, initialValue);

    case 'json':
      return normalizeJson(value, initialValue);

    case 'string':
      return normalizeString(value, initialValue);

    default:
      return value;
  }
};

var normalizeJson$1 = function normalizeJson(value) {
  if (typeof value === 'object' && 'type' in value) return JSON.stringify(value.initial);
  if (typeof value === 'object') return JSON.stringify(value);
  return JSON.stringify(value);
};

var normalizeArray1 = function normalizeArray1(value, initialValue) {
  switch (initialValue.arrayType) {
    case 'boolean':
      return normalizeBooleanArray(value, initialValue).map(function (v) {
        return v ? 'true' : 'false';
      });

    case 'number':
      return normalizeNumberArray(value, initialValue).map(function (v) {
        return v.toString();
      });

    case 'string':
      return value;

    default:
      throw new Error("We cannot support " + initialValue.arrayType + " type yet");
  }
};

var isRemoveInitialStringValue = function isRemoveInitialStringValue(value, initialValue) {
  return value === initialValue.initial;
};

var isRemoveInitialNumberValue = function isRemoveInitialNumberValue(value, initialValue) {
  return +value === initialValue.initial;
};

var isRemoveInitialBooleanValue = function isRemoveInitialBooleanValue(value, initialValue) {
  return value === "" + initialValue.initial;
};

var isRemoveInitialArrayStringValue = function isRemoveInitialArrayStringValue(value, initialValue) {
  for (var i = 0; i < initialValue.initial.length; i++) {
    if (initialValue.initial[i] !== value[i]) return false;
  }

  return true;
};

var isRemoveInitialArrayBooleanValue = function isRemoveInitialArrayBooleanValue(value, initialValue) {
  for (var i = 0; i < initialValue.initial.length; i++) {
    if (initialValue.initial[i]) {
      if (value[i] !== 'true') return false;
    } else {
      if (value[i] !== 'false') return false;
    }
  }

  return true;
};

var isRemoveInitialArrayNumberValue = function isRemoveInitialArrayNumberValue(value, initialValue) {
  for (var i = 0; i < initialValue.initial.length; i++) {
    if (initialValue.initial[i] !== +value[i]) return false;
  }

  return true;
};

var isRemoveInitialArrayValue = function isRemoveInitialArrayValue(value, initialValue) {
  if (value.length !== initialValue.initial.length) return false;

  switch (initialValue.arrayType) {
    case 'string':
      return isRemoveInitialArrayStringValue(value, initialValue);

    case 'number':
      return isRemoveInitialArrayNumberValue(value, initialValue);

    case 'boolean':
      return isRemoveInitialArrayBooleanValue(value, initialValue);
  }

  throw new Error('Unknown array type');
};

var isRemoveInitialJsonValue = function isRemoveInitialJsonValue(value, initialValue) {
  return value === JSON.stringify(initialValue.initial);
};

var isRemoveInitialValue = function isRemoveInitialValue(value, initialValue) {
  if (Array.isArray(value)) {
    if (initialValue.type === 'array') return isRemoveInitialArrayValue(value, initialValue);
    throw new Error('Array with uncorrect behavior');
  }

  switch (initialValue.type) {
    case 'string':
      return isRemoveInitialStringValue(value, initialValue);

    case 'number':
      return isRemoveInitialNumberValue(value, initialValue);

    case 'boolean':
      return isRemoveInitialBooleanValue(value, initialValue);

    case 'json':
      return isRemoveInitialJsonValue(value, initialValue);
  }

  throw new Error('Unknown initial value type');
};

var removeInitialValues = function removeInitialValues(query, context) {
  var initialValues = getInitialValues(context);

  var locationQuery = _extends({}, query);

  Object.keys(query).forEach(function (key) {
    var value = query[key];
    var initialValue = initialValues[key];
    if (!initialValue.hideIfInitial) return;

    if (isRemoveInitialValue(value, initialValue)) {
      delete locationQuery[key];
    }
  });
  return locationQuery;
};

var normalizeForLocation = function normalizeForLocation(queryValues, context) {
  var initialValues = getInitialValues(context);
  var locationValues = {};
  Object.keys(initialValues).forEach(function (key) {
    var value = queryValues[key];
    var initialValue = initialValues[key];

    if (!initialValue.active || typeof initialValue.active === 'object' && !initialValue.active.isActive) {
      return;
    }

    switch (initialValue.type) {
      case 'json':
        locationValues[key] = normalizeJson$1(value);
        break;

      case 'number':
        locationValues[key] = normalizeNumber(value, initialValue).toString();
        break;

      case 'boolean':
        locationValues[key] = normalizeBoolean(value, initialValue) ? 'true' : 'false';
        break;

      case 'string':
        locationValues[key] = normalizeString(value, initialValue);
        break;

      case 'array':
        locationValues[key] = normalizeArray1(value, initialValue);
        break;

      default:
        throw new Error('Unknown behavior error: unknown value');
    }

    var wrapper = getInitialValuesWrapper(key, context);
    wrapper.storedValue = locationValues[key];
  });
  removeUnusedQueryFields(queryValues, locationValues, context);
  locationValues = removeInitialValues(locationValues, context);
  return locationValues;
};

var removeUnusedQueryFields = function removeUnusedQueryFields(queryValues, locationValues, context) {
  if (!context.rules.removeUnusedQueryFields) {
    Object.assign(locationValues, _extends({}, queryValues, locationValues));
  }
};

var compressFromBase64 = function compressFromBase64(text) {
  return atob(text);
};

var base64Decrypt = function base64Decrypt(text) {
  return atob(text);
};

var xorEncrypt = function xorEncrypt(text, key) {
  var result = '';

  for (var i = 0; i < text.length; i++) {
    result += String.fromCharCode(text.charCodeAt(i) ^ key.charCodeAt(i % key.length));
  }

  return result;
};

var xorDecrypt = function xorDecrypt(text, key) {
  return xorEncrypt(text, key);
};

var decryptQuery = function decryptQuery(query, cryptoOptions) {
  var option = cryptoOptions;
  var decrypted;

  switch (option.method) {
    case 'base64':
      decrypted = base64Decrypt(query);
      break;

    case 'xor':
      if (option.compessedToBase64) decrypted = xorDecrypt(compressFromBase64(query), option.key);else decrypted = xorDecrypt(query, option.key);
      break;

    default:
      throw new Error("Method " + option.method + " not implemented");
  }

  var jsoned = JSON.parse(decrypted);
  return jsoned;
};

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

// Copyright Joyent, Inc. and other Node contributors.

// If obj.hasOwnProperty has been overridden, then calling
// obj.hasOwnProperty(prop) will break.
// See: https://github.com/joyent/node/issues/1707
function hasOwnProperty(obj, prop) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
}

var decode = function(qs, sep, eq, options) {
  sep = sep || '&';
  eq = eq || '=';
  var obj = {};

  if (typeof qs !== 'string' || qs.length === 0) {
    return obj;
  }

  var regexp = /\+/g;
  qs = qs.split(sep);

  var maxKeys = 1000;
  if (options && typeof options.maxKeys === 'number') {
    maxKeys = options.maxKeys;
  }

  var len = qs.length;
  // maxKeys <= 0 means that we should not limit keys count
  if (maxKeys > 0 && len > maxKeys) {
    len = maxKeys;
  }

  for (var i = 0; i < len; ++i) {
    var x = qs[i].replace(regexp, '%20'),
        idx = x.indexOf(eq),
        kstr, vstr, k, v;

    if (idx >= 0) {
      kstr = x.substr(0, idx);
      vstr = x.substr(idx + 1);
    } else {
      kstr = x;
      vstr = '';
    }

    k = decodeURIComponent(kstr);
    v = decodeURIComponent(vstr);

    if (!hasOwnProperty(obj, k)) {
      obj[k] = v;
    } else if (Array.isArray(obj[k])) {
      obj[k].push(v);
    } else {
      obj[k] = [obj[k], v];
    }
  }

  return obj;
};

// Copyright Joyent, Inc. and other Node contributors.

var stringifyPrimitive = function(v) {
  switch (typeof v) {
    case 'string':
      return v;

    case 'boolean':
      return v ? 'true' : 'false';

    case 'number':
      return isFinite(v) ? v : '';

    default:
      return '';
  }
};

var encode = function(obj, sep, eq, name) {
  sep = sep || '&';
  eq = eq || '=';
  if (obj === null) {
    obj = undefined;
  }

  if (typeof obj === 'object') {
    return Object.keys(obj).map(function(k) {
      var ks = encodeURIComponent(stringifyPrimitive(k)) + eq;
      if (Array.isArray(obj[k])) {
        return obj[k].map(function(v) {
          return ks + encodeURIComponent(stringifyPrimitive(v));
        }).join(sep);
      } else {
        return ks + encodeURIComponent(stringifyPrimitive(obj[k]));
      }
    }).join(sep);

  }

  if (!name) return '';
  return encodeURIComponent(stringifyPrimitive(name)) + eq +
         encodeURIComponent(stringifyPrimitive(obj));
};

var Q__projects_reactLocationQuery_node_modules_querystring = createCommonjsModule(function (module, exports) {

exports.decode = exports.parse = decode;
exports.encode = exports.stringify = encode;
});

var history;
var setHistory = function setHistory(h) {
  return history = h;
};
var getHistory = function getHistory() {
  return history;
};
var getLocation = function getLocation() {
  return history.location;
};

var readQuery = function readQuery(crypto) {
  var location = getLocation();
  return parseQuery(location.search, crypto);
};

var parseQuery = function parseQuery(query, crypto) {
  try {
    var q = Q__projects_reactLocationQuery_node_modules_querystring.parse(query === '' ? '' : query.substring(1));

    if (crypto.method !== 'none') {
      if (q.q) return decryptQuery(q.q, crypto);
      return {};
    }

    return q;
  } catch (e) {
    return {};
  }
};

var compressWithBase64 = function compressWithBase64(text) {
  return btoa(text);
};

var base64Encrypt = function base64Encrypt(text) {
  return btoa(text);
};

var encryptQuery = function encryptQuery(query, crypto) {
  var jsoned = JSON.stringify(query);
  var encrypted = '';

  switch (crypto.method) {
    case 'base64':
      encrypted = base64Encrypt(jsoned);
      break;

    case 'xor':
      if (crypto.compessedToBase64) encrypted = compressWithBase64(xorEncrypt(jsoned, crypto.key));else encrypted = xorEncrypt(jsoned, crypto.key);
      break;

    default:
      throw new Error("Method " + crypto.method + " not implemented");
  }

  return encrypted;
};

var stringifyQuery = function stringifyQuery(query, crypto) {
  prepareQuery(query);

  if (crypto.method !== 'none') {
    var encryptedQuery = encryptQuery(query, crypto);
    var q = Q__projects_reactLocationQuery_node_modules_querystring.stringify({
      q: encryptedQuery
    });
    return q;
  }

  return Q__projects_reactLocationQuery_node_modules_querystring.stringify(query);
};

var prepareQuery = function prepareQuery(query) {
  Object.keys(query).forEach(function (key) {
    var value = query[key];

    if (typeof value === 'object' && !Array.isArray(value)) {
      query[key] = JSON.stringify(value);
    }
  });
};

var sortByIndex = function sortByIndex(defaultValuesWrappers) {
  return function (a, b) {
    var indexA = defaultValuesWrappers[a];
    var indexB = defaultValuesWrappers[b];
    if (!indexA || !indexB) return 0;
    return indexA.index - indexB.index;
  };
};

var sortByAlphabet = function sortByAlphabet(a, b) {
  if (a === b) return 0;
  return a > b ? 1 : -1;
};

var sortByFieldLength = function sortByFieldLength(a, b) {
  return a.length - b.length;
};

var sortFieldsInQuery = function sortFieldsInQuery(query, context) {
  var result = {};
  var sortedKeys = sortBy(Object.keys(query), context);
  (context.sortOptions.sortOrder === 'asc' ? sortedKeys : sortedKeys.reverse()).forEach(function (key) {
    result[key] = query[key];
  });
  return result;
};

var sortBy = function sortBy(queryKeys, context) {
  switch (context.sortOptions.sortBy) {
    case 'order':
      return queryKeys.sort(sortByIndex(getInitialValuesWrappers(context)));

    case 'alphabet':
      return queryKeys.sort(sortByAlphabet);

    case 'fieldLength':
      return queryKeys.sort(sortByFieldLength);
  }
};

var writeQuery = function writeQuery(query, context, actionOnChange, force) {
  if (force === void 0) {
    force = false;
  }

  var sorted = sortFieldsInQuery(query, context);
  if (stringifyQuery(sorted, context.cryptoOptions) === stringifyQuery(readQuery(context.cryptoOptions), context.cryptoOptions) && !force) return;
  var preparedQuery = stringifyQuery(sorted, context.cryptoOptions);
  var history = getHistory();
  var location = getLocation();
  var action = actionOnChange === 'Push' ? history.push : history.replace;

  if (Object.keys(sorted).length === 0) {
    action(location.pathname);
  } else {
    action(location.pathname + "?" + preparedQuery);
  }
};

var calculateLocationPath = function calculateLocationPath(context, actionOnChange, force) {
  if (force === void 0) {
    force = false;
  }

  var queryValues = readQuery(context.cryptoOptions);
  var normalizedQuery = normalizeForLocation(_extends({}, queryValues, context.query), context);
  writeQuery(normalizedQuery, context, actionOnChange, force);
};
var setQueryFieldValue = function setQueryFieldValue(field, value, context, actionOnChange) {
  var wrapper = getInitialValuesWrapper(field, context);
  if (!wrapper) return;
  var initialValue = wrapper.initialValue;
  if (!initialValue.active) return;
  var active = initialValue.active;
  if (typeof active === 'object' && !active.isActive && !active.canChangeValue) return;
  context.query[field] = value;
  calculateLocationPath(context, actionOnChange, true);
};

var checkInitialValueString = function checkInitialValueString(initialValue) {
  if (typeof initialValue.initial !== 'string') throw new Error('Initial value must be a string');

  if (initialValue["enum"] !== undefined) {
    if (Array.isArray(initialValue["enum"])) {
      if (initialValue["enum"].length === 0) throw new Error('Enum array cannot be empty');
      if (!initialValue["enum"].includes(initialValue.initial)) throw new Error("Initial value is not exists in enum array, expected: " + initialValue["enum"] + ", got: " + initialValue.initial);
    } else {
      if (initialValue["enum"].array.length === 0) throw new Error('Enum array cannot be empty');
      if (!initialValue["enum"].array.includes(initialValue.initial)) throw new Error("Initial value is not exists in enum.array, expected: " + initialValue["enum"].array + ", got: " + initialValue.initial);
    }
  }
};

var checkInitialValueBoolean = function checkInitialValueBoolean(initialValue) {
  if (typeof initialValue.initial !== 'boolean') throw new Error('Initial value must be a boolean');
};

var checkInitialValueNumber = function checkInitialValueNumber(initialValue) {
  if (typeof initialValue.initial !== 'number') throw new Error('Initial value must be a number');

  if (initialValue.validate) {
    if (!initialValue.validate(initialValue.initial)) throw new Error('Initial value must be a validating by validate callback');
  }

  if (initialValue.min !== undefined) {
    if (initialValue.max !== undefined) {
      if (initialValue.min > initialValue.max) {
        throw new Error("'min' must be smaller then 'max'");
      }
    }
  }
};

var checkInitialValue = function checkInitialValue(initialValue) {
  switch (initialValue.type) {
    case 'string':
      return checkInitialValueString(initialValue);

    case 'number':
      return checkInitialValueNumber(initialValue);

    case 'boolean':
      return checkInitialValueBoolean(initialValue);
  }
};

var addInitialValue = function addInitialValue(name, initialValue, index, context) {
  checkInitialValue(initialValue);
  var wrappers = getInitialValuesWrappers(context);

  if (wrappers[name]) {
    wrappers[name].marked = true;
    wrappers[name].index = index;
    wrappers[name].initialValue = initialValue;
    wrappers[name].name = name;
  } else {
    var _extends2;

    var wrapper = {
      index: index,
      initialValue: initialValue,
      name: name,
      marked: true,
      storedValue: undefined
    };
    wrappers[name] = wrapper;
    setInitialValuesWrappers(_extends({}, context.initialValuesWrappers, (_extends2 = {}, _extends2[name] = wrapper, _extends2)), context);
  }
};

var transformInitialArray = function transformInitialArray(array) {
  checkAllItemsType(array, typeof array[0]);

  switch (typeof array[0]) {
    case 'string':
      return {
        type: 'array',
        arrayType: 'string',
        initial: array
      };

    case 'number':
      return {
        type: 'array',
        arrayType: 'number',
        initial: array
      };

    case 'boolean':
      return {
        type: 'array',
        arrayType: 'boolean',
        initial: array
      };

    default:
      throw new Error('Unsupported array type');
  }
};

var checkAllItemsType = function checkAllItemsType(array, type) {
  array.forEach(function (item) {
    if (typeof item !== type) throw new Error('All items in array must be same type');
  });
};

var transformToInitialValue = function transformToInitialValue(value, context) {
  var initialValue;

  if (typeof value === 'object') {
    if ('type' in value) {
      initialValue = _extends({}, value);
    } else {
      if (Array.isArray(value)) {
        initialValue = transformInitialArray(value);
      } else {
        throw new Error('Passed uncorrect object, please read docs');
      }
    }
  } else {
    switch (typeof value) {
      case 'string':
        initialValue = {
          type: 'string',
          initial: value
        };
        break;

      case 'number':
        initialValue = {
          type: 'number',
          initial: value
        };
        break;

      case 'boolean':
        initialValue = {
          type: 'boolean',
          initial: value
        };
        break;

      default:
        throw new Error("Unsupported initial value type, got " + typeof value);
    }
  }

  initialValue = _extends({
    hideIfInitial: context.defaultOptions.hideIfDefault,
    actionOnChange: context.defaultOptions.actionOnChange,
    active: true
  }, initialValue);
  return initialValue;
};

var Context = React.createContext(undefined);

function useLocationField(name, value) {
  var index = useIndex();
  var context = React.useContext(Context);
  if (!context) throw new Error('You must use useLocationField hook inside BrowserLocationQuery');

  if (value !== undefined) {
    var initialValue = transformToInitialValue(value, context);
    addInitialValue(name, initialValue, index, context);
    context.query[name] = normalizeValueForUser(context.query[name], initialValue, name, context);
  }

  if (context.query[name] === undefined && value === undefined) throw new Error("Unknown field: '" + name + "'");
  var initial = getInitialValueByFieldName(name, context);
  return [context.query[name], function (newValue, actionOnChange) {
    if (actionOnChange === void 0) {
      actionOnChange = initial.actionOnChange ? initial.actionOnChange : 'Replace';
    }

    return setQueryFieldValue(name, newValue, context, actionOnChange);
  }];
}

function useLocationFieldT(name) {
  return useLocationField(name);
}

function useLocationFields(values) {
  if (values === void 0) {
    values = {};
  }

  var index = useIndex();
  var context = React.useContext(Context);
  if (!context) throw new Error('You must use useLocationFields hook inside BrowserLocationQuery');
  var keys = Object.keys(values);

  if (keys.length !== 0) {
    keys.forEach(function (name) {
      var value = values[name];
      var initialValue = transformToInitialValue(value, context);
      addInitialValue(name, initialValue, index, context);
      context.query[name] = normalizeValueForUser(context.query[name], initialValue, name, context);
    });
  }

  return {
    values: keys.reduce(function (p, name) {
      return _extends({}, p, {
        name: context.query[name]
      });
    }, {}),
    setQueryField: function setQueryField(name, newValue, actionOnChange) {
      var initial = getInitialValueByFieldName(name, context);
      return setQueryFieldValue(name, newValue, context, actionOnChange ? actionOnChange : initial.actionOnChange ? initial.actionOnChange : 'Replace');
    }
  };
}

var cleanContext = function cleanContext(context) {
  context.isCleaned = true;
};

var useQueryPush = function useQueryPush(replace) {
  if (replace === void 0) {
    replace = false;
  }

  var history = reactRouterDom.useHistory();
  var context = React.useContext(Context);
  if (!context) throw new Error('You must use useQueryPush hook inside BrowserLocationQuery');

  var push = function push(path, values, innerReplace) {
    if (innerReplace === void 0) {
      innerReplace = false;
    }

    cleanContext(context);
    var action = replace || innerReplace ? history.replace : history.push;

    if (Object.keys(values).length === 0) {
      return action(path);
    }

    var query = stringifyQuery(values, context.cryptoOptions);
    return action(path + "?" + query);
  };

  return push;
};

function withLocationFields(values, values_FieldName, setQueryField_FieldName) {
  if (values_FieldName === void 0) {
    values_FieldName = 'values';
  }

  if (setQueryField_FieldName === void 0) {
    setQueryField_FieldName = 'setQueryField';
  }

  return function (WrapperComponent) {
    return function (props) {
      var _Object$assign;

      var context = React.useContext(Context);

      if (!context) {
        throw new Error('You must use withLocationFields inside BrowserLocationQuery');
      }

      var _useLocationFields = useLocationFields(values),
          _values = _useLocationFields.values,
          setQueryField = _useLocationFields.setQueryField;

      return React__default.createElement(WrapperComponent, Object.assign({}, props, (_Object$assign = {}, _Object$assign[values_FieldName] = _values, _Object$assign[setQueryField_FieldName] = setQueryField, _Object$assign)));
    };
  };
}

function withQueryPush(replaceToName) {
  return function (WrapperComponent) {
    return function (props) {
      var _Object$assign;

      var context = React.useContext(Context);

      if (!context) {
        throw new Error('You must use withQueryPush inside BrowserLocationQuery');
      }

      var name = replaceToName || 'push';
      var push = useQueryPush();
      return React__default.createElement(WrapperComponent, Object.assign({}, props, (_Object$assign = {}, _Object$assign[name] = push, _Object$assign)));
    };
  };
}

var unmarkAllInitialValuesWrappers = function unmarkAllInitialValuesWrappers(context) {
  var wrappers = getInitialValuesWrappers(context);
  Object.keys(wrappers).map(function (key) {
    return wrappers[key].marked = false;
  });
};

var deleteUnmarkedWrappers = function deleteUnmarkedWrappers(context) {
  var wrappers = getInitialValuesWrappers(context);
  setInitialValuesWrappers(Object.keys(wrappers).filter(function (key) {
    return filter(wrappers[key]);
  }).reduce(function (prev, key) {
    var _extends2;

    return _extends({}, prev, (_extends2 = {}, _extends2[key] = wrappers[key], _extends2));
  }, {}), context);
};

var filter = function filter(wrapper) {
  if (wrapper.marked === false) return false;
  if (wrapper.initialValue.active === false) return false;

  if (typeof wrapper.initialValue.active === 'object') {
    if (wrapper.initialValue.active.isActive === false) return wrapper.initialValue.active.storeValue;
  }

  return true;
};

var oldContext;

var BrowserLocationQuery = function BrowserLocationQuery(_ref) {
  var children = _ref.children,
      rules = _ref.rules,
      defaultOptions = _ref.defaultOptions,
      sortOptions = _ref.sortOptions;
  var history = reactRouterDom.useHistory();
  reactRouterDom.useLocation();
  setHistory(history);

  var newDefaultOptions = _extends({
    hideIfDefault: false,
    actionOnChange: 'Push'
  }, defaultOptions);

  var newSortOptions = _extends({
    sortBy: 'order',
    sortOrder: 'asc'
  }, sortOptions);

  var newCryptoOptions = _extends({
    method: 'none'
  }, crypto);

  var newRules = _extends({
    removeUnusedQueryFields: true
  }, rules);

  var context = oldContext && !oldContext.isCleaned ? oldContext : {
    isCleaned: false,
    query: readQuery(newCryptoOptions),
    initialValuesWrappers: {},
    defaultOptions: newDefaultOptions,
    sortOptions: newSortOptions,
    cryptoOptions: newCryptoOptions,
    rules: newRules
  };
  unmarkAllInitialValuesWrappers(context);
  oldContext = context;
  React.useEffect(function () {
    deleteUnmarkedWrappers(context);
  });
  return React__default.createElement(Context.Provider, {
    value: context
  }, React__default.createElement(InsideProvider, null, children));
};

var InsideProvider = function InsideProvider(_ref2) {
  var children = _ref2.children;
  var context = React.useContext(Context);
  React.useEffect(function () {
    calculateLocationPath(context, 'Replace');
  });
  return children;
};

var LocationQuery = function LocationQuery(_ref) {
  var _ref$values = _ref.values,
      values = _ref$values === void 0 ? {} : _ref$values,
      children = _ref.children;

  var _useLocationFields = useLocationFields(values),
      setQueryField = _useLocationFields.setQueryField;

  var context = React.useContext(Context);
  return children({
    query: context.query,
    setQueryField: setQueryField
  });
};

var Link = function Link(_ref) {
  var to = _ref.to,
      _ref$query = _ref.query,
      query = _ref$query === void 0 ? {} : _ref$query,
      props = _objectWithoutPropertiesLoose(_ref, ["to", "query"]);

  var context = React.useContext(Context);
  var newTo = to + "?" + stringifyQuery(query, context.cryptoOptions);
  return React__default.createElement(reactRouterDom.Link, Object.assign({
    to: newTo
  }, props));
};

function withLocationField(name, value, fieldName, setFieldName) {
  return function (WrapperComponent) {
    return function (props) {
      var _Object$assign;

      var context = React.useContext(Context);

      if (!context) {
        throw new Error('You must use withLocationField inside BrowserLocationQuery');
      }

      var _useLocationField = useLocationField(name, value),
          _value = _useLocationField[0],
          _setValue = _useLocationField[1];

      var _fieldName = fieldName || name;

      var _setFieldName = setFieldName || "set" + name[0].toUpperCase() + name.substring(1);

      return React__default.createElement(WrapperComponent, Object.assign({}, props, (_Object$assign = {}, _Object$assign[_fieldName] = _value, _Object$assign[_setFieldName] = _setValue, _Object$assign)));
    };
  };
}

exports.BrowserLocationQuery = BrowserLocationQuery;
exports.Link = Link;
exports.LocationQuery = LocationQuery;
exports.useLocationField = useLocationField;
exports.useLocationFieldT = useLocationFieldT;
exports.useLocationFields = useLocationFields;
exports.useQueryPush = useQueryPush;
exports.withLocationField = withLocationField;
exports.withLocationFields = withLocationFields;
exports.withQueryPush = withQueryPush;
//# sourceMappingURL=index.js.map
